# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.

import six
from . import core
from altair.utils.schemapi import Undefined
from altair.utils import parse_shorthand, parse_shorthand_plus_data


class Color(core.MarkPropFieldDefWithCondition):
    """Color schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    legend : anyOf(Legend, None)
        An object defining properties of the legend. If `null`, the legend for the encoding 
        channel will be removed.  __Default value:__ If undefined, default [legend 
        properties](https://vega.github.io/vega-lite/docs/legend.html) are applied.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 legend=Undefined, scale=Undefined, sort=Undefined, timeUnit=Undefined, **kwds):
        super(Color, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                    condition=condition, legend=legend, scale=scale, sort=sort,
                                    timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Color, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class ColorValue(core.MarkPropValueDefWithCondition):
    """ColorValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalMarkPropFieldDef, ConditionalValueDef, 
    List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(ColorValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(ColorValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Column(core.FacetFieldDef):
    """Column schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    header : Header
        An object defining properties of a facet's header.
    sort : SortOrder
        Sort order for a facet field. This can be `"ascending"`, `"descending"`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, header=Undefined,
                 sort=Undefined, timeUnit=Undefined, **kwds):
        super(Column, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                     header=header, sort=sort, timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Column, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Detail(core.FieldDef):
    """Detail schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(Detail, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                     timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Detail, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Fill(core.MarkPropFieldDefWithCondition):
    """Fill schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    legend : anyOf(Legend, None)
        An object defining properties of the legend. If `null`, the legend for the encoding 
        channel will be removed.  __Default value:__ If undefined, default [legend 
        properties](https://vega.github.io/vega-lite/docs/legend.html) are applied.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 legend=Undefined, scale=Undefined, sort=Undefined, timeUnit=Undefined, **kwds):
        super(Fill, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                   condition=condition, legend=legend, scale=scale, sort=sort,
                                   timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Fill, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class FillValue(core.MarkPropValueDefWithCondition):
    """FillValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalMarkPropFieldDef, ConditionalValueDef, 
    List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(FillValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(FillValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Href(core.FieldDefWithCondition):
    """Href schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 timeUnit=Undefined, **kwds):
        super(Href, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                   condition=condition, timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Href, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class HrefValue(core.ValueDefWithCondition):
    """HrefValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalFieldDef, ConditionalValueDef, List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(HrefValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(HrefValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Key(core.FieldDef):
    """Key schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(Key, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                  timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Key, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Latitude(core.FieldDef):
    """Latitude schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(Latitude, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                       timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Latitude, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Latitude2(core.FieldDef):
    """Latitude2 schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(Latitude2, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                        timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Latitude2, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Longitude(core.FieldDef):
    """Longitude schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(Longitude, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                        timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Longitude, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Longitude2(core.FieldDef):
    """Longitude2 schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(Longitude2, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                         timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Longitude2, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Opacity(core.MarkPropFieldDefWithCondition):
    """Opacity schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    legend : anyOf(Legend, None)
        An object defining properties of the legend. If `null`, the legend for the encoding 
        channel will be removed.  __Default value:__ If undefined, default [legend 
        properties](https://vega.github.io/vega-lite/docs/legend.html) are applied.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 legend=Undefined, scale=Undefined, sort=Undefined, timeUnit=Undefined, **kwds):
        super(Opacity, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                      condition=condition, legend=legend, scale=scale, sort=sort,
                                      timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Opacity, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class OpacityValue(core.MarkPropValueDefWithCondition):
    """OpacityValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalMarkPropFieldDef, ConditionalValueDef, 
    List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(OpacityValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(OpacityValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Order(core.OrderFieldDef):
    """Order schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    sort : SortOrder
        The sort order. One of `"ascending"` (default) or `"descending"`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, sort=Undefined,
                 timeUnit=Undefined, **kwds):
        super(Order, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin, sort=sort,
                                    timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Order, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Row(core.FacetFieldDef):
    """Row schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    header : Header
        An object defining properties of a facet's header.
    sort : SortOrder
        Sort order for a facet field. This can be `"ascending"`, `"descending"`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, header=Undefined,
                 sort=Undefined, timeUnit=Undefined, **kwds):
        super(Row, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin, header=header,
                                  sort=sort, timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Row, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Shape(core.MarkPropFieldDefWithCondition):
    """Shape schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    legend : anyOf(Legend, None)
        An object defining properties of the legend. If `null`, the legend for the encoding 
        channel will be removed.  __Default value:__ If undefined, default [legend 
        properties](https://vega.github.io/vega-lite/docs/legend.html) are applied.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 legend=Undefined, scale=Undefined, sort=Undefined, timeUnit=Undefined, **kwds):
        super(Shape, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                    condition=condition, legend=legend, scale=scale, sort=sort,
                                    timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Shape, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class ShapeValue(core.MarkPropValueDefWithCondition):
    """ShapeValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalMarkPropFieldDef, ConditionalValueDef, 
    List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(ShapeValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(ShapeValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Size(core.MarkPropFieldDefWithCondition):
    """Size schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    legend : anyOf(Legend, None)
        An object defining properties of the legend. If `null`, the legend for the encoding 
        channel will be removed.  __Default value:__ If undefined, default [legend 
        properties](https://vega.github.io/vega-lite/docs/legend.html) are applied.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 legend=Undefined, scale=Undefined, sort=Undefined, timeUnit=Undefined, **kwds):
        super(Size, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                   condition=condition, legend=legend, scale=scale, sort=sort,
                                   timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Size, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class SizeValue(core.MarkPropValueDefWithCondition):
    """SizeValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalMarkPropFieldDef, ConditionalValueDef, 
    List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(SizeValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(SizeValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Stroke(core.MarkPropFieldDefWithCondition):
    """Stroke schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    legend : anyOf(Legend, None)
        An object defining properties of the legend. If `null`, the legend for the encoding 
        channel will be removed.  __Default value:__ If undefined, default [legend 
        properties](https://vega.github.io/vega-lite/docs/legend.html) are applied.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 legend=Undefined, scale=Undefined, sort=Undefined, timeUnit=Undefined, **kwds):
        super(Stroke, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                     condition=condition, legend=legend, scale=scale, sort=sort,
                                     timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Stroke, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class StrokeValue(core.MarkPropValueDefWithCondition):
    """StrokeValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalMarkPropFieldDef, ConditionalValueDef, 
    List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(StrokeValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(StrokeValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Text(core.TextFieldDefWithCondition):
    """Text schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    format : string
        The [formatting pattern](https://vega.github.io/vega-lite/docs/format.html) for a 
        text field. If not defined, this will be determined automatically.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 format=Undefined, timeUnit=Undefined, **kwds):
        super(Text, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                   condition=condition, format=format, timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Text, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class TextValue(core.TextValueDefWithCondition):
    """TextValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalTextFieldDef, ConditionalValueDef, List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(TextValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(TextValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Tooltip(core.TextFieldDefWithCondition):
    """Tooltip schema wrapper
    
    Mapping(required=[type])
    A FieldDef with Condition<ValueDef>
    {
       condition: {value: ...},
       field: ...,
       ...
    }
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    condition : anyOf(ConditionalValueDef, List(ConditionalValueDef))
        One or more value definition(s) with a selection predicate.  __Note:__ A field 
        definition's `condition` property can only contain [value 
        definitions](https://vega.github.io/vega-lite/docs/encoding.html#value-def) since 
        Vega-Lite only allows at most one encoded field per encoding channel.
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    format : string
        The [formatting pattern](https://vega.github.io/vega-lite/docs/format.html) for a 
        text field. If not defined, this will be determined automatically.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, condition=Undefined,
                 format=Undefined, timeUnit=Undefined, **kwds):
        super(Tooltip, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                      condition=condition, format=format, timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Tooltip, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class TooltipValue(core.TextValueDefWithCondition):
    """TooltipValue schema wrapper
    
    Mapping(required=[])
    A ValueDef with Condition<ValueDef | FieldDef>
    {
       condition: {field: ...} | {value: ...},
       value: ...,
    }
    
    Attributes
    ----------
    condition : anyOf(ConditionalTextFieldDef, ConditionalValueDef, List(ConditionalValueDef))
        A field definition or one or more value definition(s) with a selection predicate.
    value : anyOf(float, string, boolean)
        A constant value in visual domain.
    """
    def __init__(self, value, condition=Undefined, **kwds):
        super(TooltipValue, self).__init__(value=value, condition=condition, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(TooltipValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class X(core.PositionFieldDef):
    """X schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    axis : anyOf(Axis, None)
        An object defining properties of axis's gridlines, ticks and labels. If `null`, the 
        axis for the encoding channel will be removed.  __Default value:__ If undefined, 
        default [axis properties](https://vega.github.io/vega-lite/docs/axis.html) are 
        applied.
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    stack : anyOf(StackOffset, None)
        Type of stacking offset if the field should be stacked. `stack` is only applicable 
        for `x` and `y` channels with continuous domains. For example, `stack` of `y` can be
         used to customize stacking for a vertical bar chart.  `stack` can be one of the 
        following values: - `"zero"`: stacking with baseline offset at zero value of the 
        scale (for creating typical stacked 
        [bar](https://vega.github.io/vega-lite/docs/stack.html#bar) and 
        [area](https://vega.github.io/vega-lite/docs/stack.html#area) chart). - 
        `"normalize"` - stacking with normalized domain (for creating [normalized stacked 
        bar and area charts](https://vega.github.io/vega-lite/docs/stack.html#normalized). 
        <br/> -`"center"` - stacking with center baseline (for 
        [streamgraph](https://vega.github.io/vega-lite/docs/stack.html#streamgraph)). - 
        `null` - No-stacking. This will produce layered 
        [bar](https://vega.github.io/vega-lite/docs/stack.html#layered-bar-chart) and area 
        chart.  __Default value:__ `zero` for plots with all of the following conditions are
         true: (1) the mark is `bar` or `area`; (2) the stacked measure channel (x or y) has
         a linear scale; (3) At least one of non-position channels mapped to an unaggregated
         field that is different from x and y.  Otherwise, `null` by default.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, axis=Undefined, bin=Undefined,
                 scale=Undefined, sort=Undefined, stack=Undefined, timeUnit=Undefined, **kwds):
        super(X, self).__init__(field=field, type=type, aggregate=aggregate, axis=axis, bin=bin,
                                scale=scale, sort=sort, stack=stack, timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(X, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class XValue(core.ValueDef):
    """XValue schema wrapper
    
    Mapping(required=[value])
    Definition object for a constant value of an encoding channel.
    
    Attributes
    ----------
    value : anyOf(float, string, boolean)
        A constant value in visual domain (e.g., `"red"` / "#0099ff" for color, values 
        between `0` to `1` for opacity).
    """
    def __init__(self, value, **kwds):
        super(XValue, self).__init__(value=value, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(XValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class X2(core.FieldDef):
    """X2 schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(X2, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                 timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(X2, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class X2Value(core.ValueDef):
    """X2Value schema wrapper
    
    Mapping(required=[value])
    Definition object for a constant value of an encoding channel.
    
    Attributes
    ----------
    value : anyOf(float, string, boolean)
        A constant value in visual domain (e.g., `"red"` / "#0099ff" for color, values 
        between `0` to `1` for opacity).
    """
    def __init__(self, value, **kwds):
        super(X2Value, self).__init__(value=value, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(X2Value, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Y(core.PositionFieldDef):
    """Y schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    axis : anyOf(Axis, None)
        An object defining properties of axis's gridlines, ticks and labels. If `null`, the 
        axis for the encoding channel will be removed.  __Default value:__ If undefined, 
        default [axis properties](https://vega.github.io/vega-lite/docs/axis.html) are 
        applied.
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    scale : anyOf(Scale, None)
        An object defining properties of the channel's scale, which is the function that 
        transforms values in the data domain (numbers, dates, strings, etc) to visual values
         (pixels, colors, sizes) of the encoding channels.  If `null`, the scale will be 
        [disabled and the data value will be directly 
        encoded](https://vega.github.io/vega-lite/docs/scale.html#disable).  __Default 
        value:__ If undefined, default [scale 
        properties](https://vega.github.io/vega-lite/docs/scale.html) are applied.
    sort : anyOf(SortOrder, SortField, None)
        Sort order for the encoded field. Supported `sort` values include `"ascending"`, 
        `"descending"` and `null` (no sorting). For fields with discrete domains, `sort` can
         also be a [sort field definition 
        object](https://vega.github.io/vega-lite/docs/sort.html#sort-field).  __Default 
        value:__ `"ascending"`
    stack : anyOf(StackOffset, None)
        Type of stacking offset if the field should be stacked. `stack` is only applicable 
        for `x` and `y` channels with continuous domains. For example, `stack` of `y` can be
         used to customize stacking for a vertical bar chart.  `stack` can be one of the 
        following values: - `"zero"`: stacking with baseline offset at zero value of the 
        scale (for creating typical stacked 
        [bar](https://vega.github.io/vega-lite/docs/stack.html#bar) and 
        [area](https://vega.github.io/vega-lite/docs/stack.html#area) chart). - 
        `"normalize"` - stacking with normalized domain (for creating [normalized stacked 
        bar and area charts](https://vega.github.io/vega-lite/docs/stack.html#normalized). 
        <br/> -`"center"` - stacking with center baseline (for 
        [streamgraph](https://vega.github.io/vega-lite/docs/stack.html#streamgraph)). - 
        `null` - No-stacking. This will produce layered 
        [bar](https://vega.github.io/vega-lite/docs/stack.html#layered-bar-chart) and area 
        chart.  __Default value:__ `zero` for plots with all of the following conditions are
         true: (1) the mark is `bar` or `area`; (2) the stacked measure channel (x or y) has
         a linear scale; (3) At least one of non-position channels mapped to an unaggregated
         field that is different from x and y.  Otherwise, `null` by default.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, axis=Undefined, bin=Undefined,
                 scale=Undefined, sort=Undefined, stack=Undefined, timeUnit=Undefined, **kwds):
        super(Y, self).__init__(field=field, type=type, aggregate=aggregate, axis=axis, bin=bin,
                                scale=scale, sort=sort, stack=stack, timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Y, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class YValue(core.ValueDef):
    """YValue schema wrapper
    
    Mapping(required=[value])
    Definition object for a constant value of an encoding channel.
    
    Attributes
    ----------
    value : anyOf(float, string, boolean)
        A constant value in visual domain (e.g., `"red"` / "#0099ff" for color, values 
        between `0` to `1` for opacity).
    """
    def __init__(self, value, **kwds):
        super(YValue, self).__init__(value=value, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(YValue, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Y2(core.FieldDef):
    """Y2 schema wrapper
    
    Mapping(required=[type])
    Definition object for a data field, its type and transformation of an encoding channel.
    
    Attributes
    ----------
    type : Type
        The encoded field's type of measurement (`"quantitative"`, `"temporal"`, 
        `"ordinal"`, or `"nominal"`). It can also be a geo type (`"latitude"`, 
        `"longitude"`, and `"geojson"`) when a [geographic 
        projection](https://vega.github.io/vega-lite/docs/projection.html) is applied.
    aggregate : Aggregate
        Aggregation function for the field (e.g., `mean`, `sum`, `median`, `min`, `max`, 
        `count`).  __Default value:__ `undefined` (None)
    bin : anyOf(boolean, BinParams)
        A flag for binning a `quantitative` field, or [an object defining binning 
        parameters](https://vega.github.io/vega-lite/docs/bin.html#params). If `true`, 
        default [binning parameters](https://vega.github.io/vega-lite/docs/bin.html) will be
         applied.  __Default value:__ `false`
    field : anyOf(string, RepeatRef)
        __Required.__ A string defining the name of the field from which to pull a data 
        value or an object defining iterated values from the 
        [`repeat`](https://vega.github.io/vega-lite/docs/repeat.html) operator.  __Note:__ 
        Dots (`.`) and brackets (`[` and `]`) can be used to access nested objects (e.g., 
        `"field": "foo.bar"` and `"field": "foo['bar']"`). If field names contain dots or 
        brackets but are not nested, you can use `\\` to escape dots and brackets (e.g., 
        `"a\\.b"` and `"a\\[0\\]"`). See more details about escaping in the [field 
        documentation](https://vega.github.io/vega-lite/docs/field.html).  __Note:__ `field`
         is not required if `aggregate` is `count`.
    timeUnit : TimeUnit
        Time unit (e.g., `year`, `yearmonth`, `month`, `hours`) for a temporal field. or [a 
        temporal field that gets casted as 
        ordinal](https://vega.github.io/vega-lite/docs/type.html#cast).  __Default value:__ 
        `undefined` (None)
    """
    _class_is_valid_at_instantiation = False

    def __init__(self, field, type=Undefined, aggregate=Undefined, bin=Undefined, timeUnit=Undefined,
                 **kwds):
        super(Y2, self).__init__(field=field, type=type, aggregate=aggregate, bin=bin,
                                 timeUnit=timeUnit, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        type_ = getattr(self, 'type', Undefined)
        context = context or {}
        if not isinstance(self.field, six.string_types):
            # field is a RepeatSpec or similar; cannot infer type
            kwds = {}
        elif type_ is Undefined and 'data' in context:
            kwds = parse_shorthand_plus_data(self.field, context['data'])
        else:
            kwds = parse_shorthand(self.field)
        self._kwds.update(kwds)
        return super(Y2, self).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)


class Y2Value(core.ValueDef):
    """Y2Value schema wrapper
    
    Mapping(required=[value])
    Definition object for a constant value of an encoding channel.
    
    Attributes
    ----------
    value : anyOf(float, string, boolean)
        A constant value in visual domain (e.g., `"red"` / "#0099ff" for color, values 
        between `0` to `1` for opacity).
    """
    def __init__(self, value, **kwds):
        super(Y2Value, self).__init__(value=value, **kwds)

    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                if 'data' in context:
                    kwds = parse_shorthand_plus_data(condition['field'], context['data'])
                else:
                    kwds = parse_shorthand(condition['field'])
                copy = self.copy()
                copy.condition.update(kwds)
        return super(Y2Value, copy).to_dict(validate=validate,
                                                ignore=ignore,
                                                context=context)



"""
Normalized Stacked Bar Chart
----------------------------
This example shows how to make a normalized stacked bar chart.
"""

import altair as alt
from altair.expr import datum, if_
from vega_datasets import data

source = data.population.url

chart = alt.Chart(source).mark_bar().encode(
    alt.X('age:O', scale=alt.Scale(rangeStep=17)),
    alt.Y('sum(people):Q',
        axis=alt.Axis(title='population'),
        stack='normalize'
    ),
    alt.Color('gender:N',
        scale=alt.Scale(range=["#EA98D2", "#659CCA"])
    )
).transform_filter(
    datum.year == 2000
).transform_calculate(
    "gender", if_(datum.sex == 2, 'Female', 'Male')
)



"""
Simple Bar Chart with Labels
============================
This example shows a basic horizontal bar chart with labels created with Altair.
"""

import altair as alt
import pandas as pd

data = pd.DataFrame({
    'a': ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'],
    'b': [28, 55, 43, 91, 81, 53, 19, 87, 52]
})

bars = alt.Chart(data).mark_bar().encode(
    y='a',
    x='b'
)

text = bars.mark_text(
    align='left',
    baseline='middle',
    dx=3
).encode(
    text='b'
)

chart = bars + text



"""
Gantt Chart
-----------------
This example shows how to make a simple gantt chart.
"""
# category: basic charts

import altair as alt
import pandas as pd

data = pd.DataFrame([
    {"task": "A","start": 1, "end": 3},
    {"task": "B","start": 3, "end": 8},
    {"task": "C","start": 8, "end": 10}
])

chart = alt.Chart(data).mark_bar().encode(
    x='start',
    x2='end',
    y='task',
)



"""
Connected Scatterplot (Lines with Custom Paths)
-----------------------------------------------

This example shows how layering can be used to build a plot. This dataset tracks miles driven per capita along with gas prices annually from 1956 to 2010. It is based on the May 2, 2010 New York Times article 'Driving Shifts Into Reverse'. See http://mbostock.github.io/protovis/ex/driving.html .
"""

import altair as alt
from vega_datasets import data

driving = data.driving()

lines = alt.Chart(driving).mark_line().encode(
    alt.X('miles', scale=alt.Scale(zero=False)),
    alt.Y('gas', scale=alt.Scale(zero=False)),
    order='year'
)

points = alt.Chart(driving).mark_circle().encode(
    alt.X('miles', scale=alt.Scale(zero=False)),
    alt.Y('gas', scale=alt.Scale(zero=False)),
    order='year'
)

chart = lines + points



import pytest
import warnings
import json

import numpy as np
import pandas as pd

from .. import parse_shorthand, parse_shorthand_plus_data, update_nested


def test_parse_shorthand():
    def check(s, **kwargs):
        assert parse_shorthand(s) == kwargs

    check('')

    # Fields alone
    check('foobar', field='foobar')
    check('blah:(fd ', field='blah:(fd ')

    # Fields with type
    check('foobar:quantitative', type='quantitative', field='foobar')
    check('foobar:nominal', type='nominal', field='foobar')
    check('foobar:ordinal', type='ordinal', field='foobar')
    check('foobar:temporal', type='temporal', field='foobar')

    check('foobar:Q', type='quantitative', field='foobar')
    check('foobar:N', type='nominal', field='foobar')
    check('foobar:O', type='ordinal', field='foobar')
    check('foobar:T', type='temporal', field='foobar')

    # Fields with aggregate and/or type
    check('average(foobar)', field='foobar', aggregate='average')
    check('min(foobar):temporal', type='temporal', field='foobar', aggregate='min')
    check('sum(foobar):Q', type='quantitative', field='foobar', aggregate='sum')

    # check that invalid arguments are not split-out
    check('invalid(blah)', field='invalid(blah)')
    check('blah:invalid', field='blah:invalid')
    check('invalid(blah):invalid', field='invalid(blah):invalid')

    # check parsing in presence of strange characters
    check('average(a b:(c\nd):Q', aggregate='average',
          field='a b:(c\nd', type='quantitative')


def test_parse_shorthand_plus_data():
    def check(s, data, **kwargs):
        assert parse_shorthand_plus_data(s, data) == kwargs

    data = pd.DataFrame({'x': [1, 2, 3, 4, 5],
                         'y': ['A', 'B', 'C', 'D', 'E'],
                         'z': pd.date_range('2018-01-01', periods=5, freq='D')})

    check('x', data, field='x', type='quantitative')
    check('y', data, field='y', type='nominal')
    check('z', data, field='z', type='temporal')
    check('count(x)', data, field='x', aggregate='count', type='quantitative')
    check('mean(*)', data, field='*', aggregate='mean')


def test_update_nested():
    original = {'x': {'b': {'foo': 2}, 'c': 4}}
    update = {'x': {'b': {'foo': 5}, 'd': 6}, 'y': 40}

    output = update_nested(original, update, copy=True)
    assert output is not original
    assert output == {'x': {'b': {'foo': 5}, 'c': 4, 'd': 6}, 'y': 40}

    output2 = update_nested(original, update)
    assert output2 is original
    assert output == output2



"""
Seattle Weather Interactive
===========================
This chart provides an interactive exploration of Seattle weather over the
course of the year. It includes a one-axis brush selection to easily
see the distribution of weather types in a particular date range.
"""
# category: interactive

import altair as alt
from vega_datasets import data

scale = alt.Scale(domain=['sun', 'fog', 'drizzle', 'rain', 'snow'],
                  range=['#e7ba52', '#a7a7a7', '#aec7e8', '#1f77b4', '#9467bd'])
color = alt.Color('weather:N', scale=scale)

# We create two selections:
# - a brush that is active on the top panel
# - a multi-click that is active on the bottom panel
brush = alt.selection_interval(encodings=['x'])
click = alt.selection_multi(encodings=['color'])

# Top panel is scatter plot of temperature vs time
points = alt.Chart().mark_point().encode(
    alt.X('date:T', timeUnit='monthdate', axis=alt.Axis(title='Date')),
    alt.Y('temp_max:Q',
        axis=alt.Axis(title='Maximum Daily Temperature (C)'),
        scale=alt.Scale(domain=[-5, 40])
    ),
    color=alt.condition(brush, color, alt.value('lightgray')),
    size=alt.Size('precipitation:Q', scale=alt.Scale(range=[5, 200]))
).properties(
    width=600,
    height=300,
    selection=brush
).transform_filter(
    click.ref()
)

# Bottom panel is a bar chart of weather type
bars = alt.Chart().mark_bar().encode(
    x='count(*):Q',
    y='weather:N',
    color=alt.condition(click, color, alt.value('lightgray')),
).transform_filter(
    brush.ref()
).properties(
    width=600,
    selection=click
)

chart = alt.vconcat(points, bars,
    data=data.seattle_weather.url,
    title="Seattle Weather: 2012-2015"
)



import json

from . import urls


def to_html(json_dict,
            template=None, title=None,
            d3_js_url=urls.D3_JS_URL,
            vega_js_url=urls.VEGA_JS_URL,
            vegalite_js_url=urls.VEGALITE_JS_URL,
            vegaembed_js_url=urls.VEGAEMBED_JS_URL,
            **kwargs):
    """Embed a Vega-Lite JSON into an HTML document.

    Parameters
    ----------
    json_dict : dict
        A dictionary describing the Vega-Lite specification.
    template : string
        The HTML template to use. This should have a format method, which
        accepts a "spec" and "title" argument. Note that a standard Python
        format string meets these requirements.
        By default, uses DEFAULT_TEMPLATE.
    title: string
        The title to use in the document. Default is "Vega-Lite Chart"

    Returns
    -------
    html : string
        A string of HTML representing the chart
    """
    if template is None:
        template = DEFAULT_TEMPLATE
    if title is None:
        title = "Vega-Lite Chart"
    spec = json.dumps(json_dict, indent=4)
    return template.format(spec=spec, title=title,
                           d3_js_url=d3_js_url,
                           vega_js_url=vega_js_url,
                           vegalite_js_url=vegalite_js_url,
                           vegaembed_js_url=vegaembed_js_url,
                           **kwargs)


DEFAULT_TEMPLATE = """
<!DOCTYPE html>
<head>
  <title>{title}</title>
  <meta charset="utf-8">

  <script src="{d3_js_url}"></script>
  <script src="{vega_js_url}"></script>
  <script src="{vegalite_js_url}"></script>
  <script src="{vegaembed_js_url}" charset="utf-8"></script>

  <style media="screen">
    /* Add space between vega-embed links  */
    .vega-actions a {{
      margin-right: 5px;
    }}
  </style>
</head>
<body>
  <!-- Container for the visualization -->
  <div id="vis"></div>

  <script>
  var vlSpec = {spec}

  var embedSpec = {{
    mode: "vega-lite",  // Instruct Vega-Embed to use the Vega-Lite compiler
    spec: vlSpec
  }};

  // Embed the visualization in the container with id `vis`
  vg.embed("#vis", embedSpec, function(error, result) {{
    // Callback receiving the View instance and parsed Vega spec
    // result.view is the View, which resides under the '#vis' element
  }});
  </script>
</body>
</html>
"""



"""
Error Bars showing Confidence Interval
======================================
This example shows how to show error bars using covidence intervals
"""

import altair as alt
from vega_datasets import data

barley = data.barley()

points = alt.Chart(barley).mark_point(filled=True).encode(
    alt.X(
        'mean(yield)',
        scale=alt.Scale(zero=False),
        axis=alt.Axis(title='Barley Yield')
    ),
    y='variety',
    color=alt.value('black')
)

error_bars = alt.Chart(barley).mark_rule().encode(
    x='ci0(yield)',
    x2='ci1(yield)',
    y='variety'
)

chart = points + error_bars



"""
Table Binned Heatmap
-----------------------
This example shows how to make a heatmap.
"""
# category: basic charts

import altair as alt
from vega_datasets import data

source = data.movies.url

chart = alt.Chart(source).mark_rect().encode(
    alt.X('IMDB_Rating:Q', bin=alt.Bin(maxbins=60)),
    alt.Y('Rotten_Tomatoes_Rating:Q', bin=alt.Bin(maxbins=40)),
    alt.Color('count(IMDB_Rating):Q', scale=alt.Scale(scheme='greenblue'))
)



from __future__ import division

import operator

import pytest

from ... import expr
from .. import datum


def test_unary_operations():
    OP_MAP = {'-': operator.neg, '+': operator.pos}
    for op, func in OP_MAP.items():
        z = func(datum.xxx)
        assert repr(z) == '({0}datum.xxx)'.format(op)


def test_binary_operations():
    OP_MAP = {'+': operator.add, '-': operator.sub,
              '*': operator.mul, '/': operator.floordiv,
              '%': operator.mod, '===': operator.eq,
              '<': operator.lt, '<=': operator.le,
              '>': operator.gt, '>=': operator.ge,
              '!==': operator.ne, '&&': operator.and_,
              '||': operator.or_}
    # When these are on the RHS, the opposite is evaluated instead.
    INEQ_REVERSE = {'>':'<', '<': '>',
                    '<=':'>=', '>=':'<=',
                    '===':'===', '!==':'!=='}
    for op, func in OP_MAP.items():
        z1 = func(datum.xxx, 2)
        assert repr(z1) == '(datum.xxx {0} 2)'.format(op)

        z2 = func(2, datum.xxx)
        if op in INEQ_REVERSE:
            assert repr(z2) == '(datum.xxx {0} 2)'.format(INEQ_REVERSE[op])
        else:
            assert repr(z2) == '(2 {0} datum.xxx)'.format(op)

        z3 = func(datum.xxx, datum.yyy)
        assert repr(z3) == '(datum.xxx {0} datum.yyy)'.format(op)


def test_abs():
    z = abs(datum.xxx)
    assert repr(z) == 'abs(datum.xxx)'


def test_expr_funcs():
    """test all functions defined in expr.funcs"""
    name_map = {val:key for key, val in expr.funcs.NAME_MAP.items()}
    for funcname in expr.funcs.__all__:
        func = getattr(expr, funcname)
        z = func(datum.xxx)
        assert repr(z) == '{0}(datum.xxx)'.format(name_map.get(funcname,
                                                               funcname))


def test_expr_consts():
    """Test all constants defined in expr.consts"""
    name_map = {val:key for key, val in expr.consts.NAME_MAP.items()}
    for constname in expr.consts.__all__:
        const = getattr(expr, constname)
        z = const * datum.xxx
        assert repr(z) == '({0} * datum.xxx)'.format(name_map.get(constname,
                                                                constname))


def test_json_reprs():
    """Test JSON representations of special values"""
    assert repr(datum.xxx == None) == '(datum.xxx === null)'
    assert repr(datum.xxx == False) == '(datum.xxx === false)'
    assert repr(datum.xxx == True) == '(datum.xxx === true)'


def test_to_dict():
    ex = (datum.xxx * 2 > datum.yyy)
    assert ex.to_dict() == repr(ex)


def test_copy():
    ex = datum.xxx * 2 > abs(datum.yyy)
    ex_copy = ex.copy()
    assert ex.to_dict() == ex_copy.to_dict()


def test_datum_getattr():
    x = datum['foo']
    assert repr(x) == "datum['foo']"



"""
Horizon Graph
-------------
This example shows how to make a Horizon Graph with 2 layers. (See https://idl.cs.washington.edu/papers/horizon/ for more details on Horizon Graphs.)
"""

import altair as alt
from altair.expr import datum
import pandas as pd

df = pd.DataFrame([
      {"x": 1,  "y": 28}, {"x": 2,  "y": 55},
      {"x": 3,  "y": 43}, {"x": 4,  "y": 91},
      {"x": 5,  "y": 81}, {"x": 6,  "y": 53},
      {"x": 7,  "y": 19}, {"x": 8,  "y": 87},
      {"x": 9,  "y": 52}, {"x": 10, "y": 48},
      {"x": 11, "y": 24}, {"x": 12, "y": 49},
      {"x": 13, "y": 87}, {"x": 14, "y": 66},
      {"x": 15, "y": 17}, {"x": 16, "y": 27},
      {"x": 17, "y": 68}, {"x": 18, "y": 16},
      {"x": 19, "y": 49}, {"x": 20, "y": 15}
    ])

area1 = alt.Chart(df).mark_area(
    clip=True,
    interpolate='monotone'
).encode(
    alt.X('x', scale=alt.Scale(zero=False, nice=False)),
    alt.Y('y', scale=alt.Scale(domain=[0, 50]), axis=alt.Axis(title='y')),
    opacity = alt.value(0.6)
).properties(
    width=300,
    height=50
)

area2 = area1.encode(
    y='ny:Q'
).transform_calculate(
    "ny", datum.y - 50
)

chart = area1 + area2



"""
Gapminder Bubble Plot
=====================
This example shows how to make a bubble plot showing the correlation between health and income for 187 countries in the world (modified from an example in Lisa Charlotte Rost's blog post 'One Chart, Twelve Charting Libraries' --http://lisacharlotterost.github.io/2016/05/17/one-chart-code/).
"""
# category: basic charts

import altair as alt
from vega_datasets import data

gapminder = data.gapminder_health_income.url

chart = alt.Chart(gapminder).mark_circle().encode(
    alt.X('income:Q', scale=alt.Scale(type='log')),
    alt.Y('health:Q', scale=alt.Scale(zero=False)),
    size='population:Q'
)



"""
Line Chart with Layered Aggregates
----------------------------------
This example shows how to make a multi series line chart of the daily closing
stock prices for AAPL, AMZN, GOOG, IBM, and MSFT between 2000 and 2010, along
with a layered rule showing the average values.
"""

import altair as alt
from vega_datasets import data

stocks = data.stocks()

line = alt.Chart(stocks).mark_line().encode(
    x='date',
    y='price',
    color='symbol'
).properties(
    width=600,
    title="Daily closing prices with their aggregate prices"
).interactive(bind_y=False)

rule = alt.Chart(stocks).mark_rule().encode(
    alt.Y('average(price)'),
    color='symbol',
    size=alt.value(2)
)

chart = line + rule



"""
Steam and Leaf Plot
-------------------
This example shows how to make a steam and leaf plot.
"""

import altair as alt
import pandas as pd
import numpy as np
np.random.seed(42)

# Generating random data
original_data = pd.DataFrame({'samples':np.array(np.random.normal(50, 15, 100), dtype=np.int)})

# Splitting steam and leaf
original_data['stem'] = original_data['samples'].apply(lambda x: str(x)[:-1])
original_data['leaf'] = original_data['samples'].apply(lambda x: str(x)[-1])

original_data.sort_values(by=['stem', 'leaf'], inplace=True)
original_data.reset_index(inplace=True, drop=True)

# Determining leaf position
get_position = lambda x: 1 + pd.Series(range(len(x)))

original_data['position'] = original_data.groupby('stem')\
                                         .apply(get_position)\
                                         .reset_index(drop=True)

# Creating stem and leaf plot
chart = alt.Chart(original_data).mark_text(
    align='left',
    baseline='middle',
    dx=-5
).encode(
    alt.X('position:Q',
        axis=alt.Axis(title='', ticks=False,labels=False,grid=False)
    ),
    alt.Y('stem:N', axis=alt.Axis(title='', tickSize=0)),
    text = 'leaf:N'
).configure_axis(
    labelFontSize=20
).configure_text(
    fontSize=20
)



"""
Layered Plot with Dual-Axis
---------------------------
This example shows how to combine two plots and keep their axes.
"""

import altair as alt
from vega_datasets import data

source = data.seattle_weather()

base = alt.Chart(source).encode(
    alt.X('date:O',
        axis=alt.Axis(format='%b'),
        timeUnit='month',
        scale=alt.Scale(zero=False)
    )
)

bar = base.mark_bar().encode(
    y='mean(precipitation)'
)


line =  base.mark_line(color='red').encode(
    y='mean(temp_max)',
)

chart = (bar + line).resolve_scale(y='independent')



"""
Diverging Stacked Bar Chart
---------------------------
This example shows a diverging stacked bar chart for sentiments towards a set of eight questions, displayed as percentages with neutral responses straddling the 0% mark.
"""

import altair as alt

data =     [
      {
        "question": "Question 1",
        "type": "Strongly disagree",
        "value": 24,
        "percentage": 0.7,
        "percentage_start": -19.1,
        "percentage_end": -18.4
      },
      {
        "question": "Question 1",
        "type": "Disagree",
        "value": 294,
        "percentage": 9.1,
        "percentage_start": -18.4,
        "percentage_end": -9.2
      },
      {
        "question": "Question 1",
        "type": "Neither agree nor disagree",
        "value": 594,
        "percentage": 18.5,
        "percentage_start": -9.2,
        "percentage_end": 9.2
      },
      {
        "question": "Question 1",
        "type": "Agree",
        "value": 1927,
        "percentage": 59.9,
        "percentage_start": 9.2,
        "percentage_end": 69.2
      },
      {
        "question": "Question 1",
        "type": "Strongly agree",
        "value": 376,
        "percentage": 11.7,
        "percentage_start": 69.2,
        "percentage_end": 80.9
      },

      {
        "question": "Question 2",
        "type": "Strongly disagree",
        "value": 2,
        "percentage": 18.2,
        "percentage_start": -36.4,
        "percentage_end": -18.2
      },
      {
        "question": "Question 2",
        "type": "Disagree",
        "value": 2,
        "percentage": 18.2,
        "percentage_start": -18.2,
        "percentage_end": 0
      },
      {
        "question": "Question 2",
        "type": "Neither agree nor disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": 0,
        "percentage_end": 0
      },
      {
        "question": "Question 2",
        "type": "Agree",
        "value": 7,
        "percentage": 63.6,
        "percentage_start": 0,
        "percentage_end": 63.6
      },
      {
        "question": "Question 2",
        "type": "Strongly agree",
        "value": 11,
        "percentage": 0,
        "percentage_start": 63.6,
        "percentage_end": 63.6
      },

      {
        "question": "Question 3",
        "type": "Strongly disagree",
        "value": 2,
        "percentage": 20,
        "percentage_start": -30,
        "percentage_end": -10
      },
      {
        "question": "Question 3",
        "type": "Disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": -10,
        "percentage_end": -10
      },
      {
        "question": "Question 3",
        "type": "Neither agree nor disagree",
        "value": 2,
        "percentage": 20,
        "percentage_start": -10,
        "percentage_end": 10
      },
      {
        "question": "Question 3",
        "type": "Agree",
        "value": 4,
        "percentage": 40,
        "percentage_start": 10,
        "percentage_end": 50
      },
      {
        "question": "Question 3",
        "type": "Strongly agree",
        "value": 2,
        "percentage": 20,
        "percentage_start": 50,
        "percentage_end": 70
      },

      {
        "question": "Question 4",
        "type": "Strongly disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": -15.6,
        "percentage_end": -15.6
      },
      {
        "question": "Question 4",
        "type": "Disagree",
        "value": 2,
        "percentage": 12.5,
        "percentage_start": -15.6,
        "percentage_end": -3.1
      },
      {
        "question": "Question 4",
        "type": "Neither agree nor disagree",
        "value": 1,
        "percentage": 6.3,
        "percentage_start": -3.1,
        "percentage_end": 3.1
      },
      {
        "question": "Question 4",
        "type": "Agree",
        "value": 7,
        "percentage": 43.8,
        "percentage_start": 3.1,
        "percentage_end": 46.9
      },
      {
        "question": "Question 4",
        "type": "Strongly agree",
        "value": 6,
        "percentage": 37.5,
        "percentage_start": 46.9,
        "percentage_end": 84.4
      },

      {
        "question": "Question 5",
        "type": "Strongly disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": -10.4,
        "percentage_end": -10.4
      },
      {
        "question": "Question 5",
        "type": "Disagree",
        "value": 1,
        "percentage": 4.2,
        "percentage_start": -10.4,
        "percentage_end": -6.3
      },
      {
        "question": "Question 5",
        "type": "Neither agree nor disagree",
        "value": 3,
        "percentage": 12.5,
        "percentage_start": -6.3,
        "percentage_end": 6.3
      },
      {
        "question": "Question 5",
        "type": "Agree",
        "value": 16,
        "percentage": 66.7,
        "percentage_start": 6.3,
        "percentage_end": 72.9
      },
      {
        "question": "Question 5",
        "type": "Strongly agree",
        "value": 4,
        "percentage": 16.7,
        "percentage_start": 72.9,
        "percentage_end": 89.6
      },

      {
        "question": "Question 6",
        "type": "Strongly disagree",
        "value": 1,
        "percentage": 6.3,
        "percentage_start": -18.8,
        "percentage_end": -12.5
      },
      {
        "question": "Question 6",
        "type": "Disagree",
        "value": 1,
        "percentage": 6.3,
        "percentage_start": -12.5,
        "percentage_end": -6.3
      },
      {
        "question": "Question 6",
        "type": "Neither agree nor disagree",
        "value": 2,
        "percentage": 12.5,
        "percentage_start": -6.3,
        "percentage_end": 6.3
      },
      {
        "question": "Question 6",
        "type": "Agree",
        "value": 9,
        "percentage": 56.3,
        "percentage_start": 6.3,
        "percentage_end": 62.5
      },
      {
        "question": "Question 6",
        "type": "Strongly agree",
        "value": 3,
        "percentage": 18.8,
        "percentage_start": 62.5,
        "percentage_end": 81.3
      },

      {
        "question": "Question 7",
        "type": "Strongly disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": -10,
        "percentage_end": -10
      },
      {
        "question": "Question 7",
        "type": "Disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": -10,
        "percentage_end": -10
      },
      {
        "question": "Question 7",
        "type": "Neither agree nor disagree",
        "value": 1,
        "percentage": 20,
        "percentage_start": -10,
        "percentage_end": 10
      },
      {
        "question": "Question 7",
        "type": "Agree",
        "value": 4,
        "percentage": 80,
        "percentage_start": 10,
        "percentage_end": 90
      },
      {
        "question": "Question 7",
        "type": "Strongly agree",
        "value": 0,
        "percentage": 0,
        "percentage_start": 90,
        "percentage_end": 90
      },

      {
        "question": "Question 8",
        "type": "Strongly disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": 0,
        "percentage_end": 0
      },
      {
        "question": "Question 8",
        "type": "Disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": 0,
        "percentage_end": 0
      },
      {
        "question": "Question 8",
        "type": "Neither agree nor disagree",
        "value": 0,
        "percentage": 0,
        "percentage_start": 0,
        "percentage_end": 0
      },
      {
        "question": "Question 8",
        "type": "Agree",
        "value": 0,
        "percentage": 0,
        "percentage_start": 0,
        "percentage_end": 0
      },
      {
        "question": "Question 8",
        "type": "Strongly agree",
        "value": 2,
        "percentage": 100,
        "percentage_start": 0,
        "percentage_end": 100
      }
    ]


color_scale = alt.Scale(
            domain=["Strongly disagree",
            "Disagree",
            "Neither agree nor disagree",
            "Agree",
            "Strongly agree"],
            range=["#c30d24", "#f3a583", "#cccccc", "#94c6da", "#1770ab"]
        )

y_axis = alt.Axis(title = 'Question',
                  offset = 5,
                  ticks = False,
                  minExtent = 60,
                  domain = False)

source = alt.pd.DataFrame(data)

chart = alt.Chart(source).mark_bar().encode(
    x='percentage_start:Q',
    x2='percentage_end:Q',
    y=alt.Y('question:N', axis=y_axis),
    color=alt.Color(
        'type:N',
        legend=alt.Legend( title='Response'),
        scale = color_scale,
    )
)



"""
Layered Bar Chart with Line as Mean
-----------------------------------
This example shows mean overlay over precipitation chart.
"""

import altair as alt
from vega_datasets import data

source = data.seattle_weather()

bar = alt.Chart(source).mark_bar().encode(
    alt.X('date:O', timeUnit='month'),
    alt.Y('mean(precipitation):Q')
)

rule = alt.Chart(source).mark_rule(color = 'red').encode(
    y='mean(precipitation)',
    size=alt.value(3)
)

chart = bar + rule



"""
Strip Plot
-----------------
This example shows the relationship between horsepower and the numbver of cylinders using tick marks.
"""
# category: basic charts

import altair as alt
from vega_datasets import data

source = data.cars()

chart = alt.Chart(source).mark_tick().encode(
    x='Horsepower:Q',
    y='Cylinders:O'
)



"""
Trellis Stacked Bar Chart
=========================
This is an example of a horizontal stacked bar chart using data which contains crop yields over different regions and different years in the 1930s.
"""

import altair as alt
from vega_datasets import data

barley = data.barley()

chart = alt.Chart(barley).mark_bar().encode(
    column='year',
    x='sum(yield)',
    y='variety',
    color='site'
).properties(
    width=250
)



"""
Interactive Average
===================
The plot below uses an interval selection, which causes the chart to include an interactive brush
(shown in grey). The brush selection parameterizes the red guideline, which visualizes the average
value within the selected interval.
"""
# category: interactive

import altair as alt
from vega_datasets import data

weather = data.seattle_weather.url
brush = alt.selection(type='interval', encodings=['x'])

bars = alt.Chart().mark_bar().encode(
    alt.X('date:O', timeUnit='month'),
    y='mean(precipitation):Q',
    opacity=alt.condition(brush, alt.OpacityValue(1), alt.OpacityValue(0.7))
).properties(
    selection=brush
)

line = alt.Chart().mark_rule(color='firebrick').encode(
    y='mean(precipitation):Q',
    size=alt.SizeValue(3)
).transform_filter(brush.ref())

chart = alt.layer(bars + line, data=weather)



"""
Layering text over a heatmap
----------------------------

An example of a layered chart of text over a heatmap using the cars dataset.
"""

import altair as alt
from vega_datasets import data

cars = data.cars.url

heatmap = alt.Chart(cars).mark_rect().encode(
    alt.X('Cylinders:O', scale=alt.Scale(paddingInner=0)),
    alt.Y('Origin:O', scale=alt.Scale(paddingInner=0)),
    alt.Color('count(*):Q')
)

text = alt.Chart(cars).mark_text(
    baseline='middle'
).encode(
    alt.X('Cylinders:O'),
    alt.Y('Origin:O'),
    alt.Text('count(*):Q'),
    color=alt.condition("datum['count_*'] > 100",
                        alt.value('black'),
                        alt.value('white'))
)

chart = heatmap + text



import functools
import json
import os
import random
import uuid

import pandas as pd
from toolz.curried import curry, pipe

from ..utils.core import sanitize_dataframe


#==============================================================================
# Data model transformers
#
# A data model transformer is a pure function that takes a dict or DataFrame
# and returns a transformed version of a dict or DataFrame. The dict objects
# will be the Data portion of the VegaLite schema. The idea is that user can
# pipe a sequence of these data transformers together to prepare the data before
# it hits the renderer.
#
# In this version of Altair, renderers only deal with the dict form of a
# VegaLite spec, after the Data model has been put into a schema compliant
# form.
#
# A data model transformer has the following type signature:
# DataModelType = Union[dict, pd.DataFrame]
# DataModelTransformerType = Callable[[DataModelType, KwArgs], DataModelType]
#==============================================================================


class MaxRowsError(Exception):
    """Raised when a data model has too many rows."""
    pass


@curry
def limit_rows(data, max_rows=5000):
    """Raise MaxRowsError if the data model has more than max_rows."""
    if not isinstance(data, (list, pd.DataFrame)):
        raise TypeError('Expected dict or DataFrame, got: {}'.format(type(data)))
    if len(data) > max_rows:
        raise MaxRowsError('The number of rows in your dataset is greater than the max of {}'.format(max_rows))
    return data


@curry
def sample(data, n=None, frac=None):
    """Reduce the size of the data model by sampling without replacement."""
    _check_data_type(data)
    if isinstance(data, pd.DataFrame):
        return data.sample(n=n, frac=frac)
    elif isinstance(data, dict):
        if 'values' in data:
            values = data['values']
            n = n if n else int(frac*len(values))
            values = random.sample(values, n)
            return {'values': values}


@curry
def to_json(data, filename=None, prefix='altair-data', base_url='/', nbserver_cwd='~'):
    """Write the data model to a .json file and return a url based data model."""
    _check_data_type(data)
    ext = '.json'
    filename, url = _compute_filename_and_url(filename=filename, prefix=prefix,
                                         base_url=base_url, nbserver_cwd=nbserver_cwd,
                                         ext=ext)
    if isinstance(data, pd.DataFrame):
        data = sanitize_dataframe(data)
        data.to_json(filename, orient='records')
    elif isinstance(data, dict):
        if 'values' not in data:
            raise KeyError('values expected in data dict, but not present.')
        values = data['values']
        with open(filename) as f:
            json.dump(values, f)
    return {
        'url': url,
        'format': {'type': 'json'}
    }


@curry
def to_csv(data, filename=None, prefix='altair-data', base_url='/', nbserver_cwd='~'):
    """Write the data model to a .csv file and return a url based data model."""
    _check_data_type(data)
    ext = '.csv'
    filename, url = _compute_filename_and_url(filename=filename, prefix=prefix,
                                         base_url=base_url, nbserver_cwd=nbserver_cwd,
                                         ext=ext)
    if isinstance(data, pd.DataFrame):
        data = sanitize_dataframe(data)
        data.to_csv(filename)
        return {
            'url': url,
            'format': {'type': 'csv'}
        }
    elif isinstance(data, dict):
        raise NotImplementedError('to_csv only works with Pandas DataFrame objects.')


@curry
def to_values(data):
    """Replace a DataFrame by a data model with values."""
    _check_data_type(data)
    if isinstance(data, pd.DataFrame):
        data = sanitize_dataframe(data)
        return {'values': data.to_dict(orient='records')}
    elif isinstance(data, dict):
        if 'values' not in data:
            raise KeyError('values expected in data dict, but not present.')
        return data


@curry
def default_data_transformer(data):
    return pipe(data, limit_rows, to_values)


#==============================================================================
# Private utilities
#==============================================================================


def _check_data_type(data):
    """Raise if the data is not a dict or DataFrame."""
    if not isinstance(data, (dict, pd.DataFrame)):
        raise TypeError('Expected dict or DataFrame, got: {}'.format(type(data)))


def _url_path_join(*pieces):
    """Join components of url into a relative url
    Use to prevent double slash when joining subpath. This will leave the
    initial and final / in place
    """
    initial = pieces[0].startswith('/')
    final = pieces[-1].endswith('/')
    stripped = [s.strip('/') for s in pieces]
    result = '/'.join(s for s in stripped if s)
    if initial: result = '/' + result
    if final: result = result + '/'
    if result == '//': result = '/'
    return result


def _compute_nbserver_url(filename, base_url='/', nbserver_cwd='~'):
    nbserver_cwd = os.path.expanduser(nbserver_cwd)
    here = os.getcwd()
    rel_path = os.path.relpath(here, nbserver_cwd)
    return _url_path_join(base_url, '/files/', rel_path, filename)


def _compute_uuid_filename(prefix, ext):
    return prefix + '-' + str(uuid.uuid4()) + ext


def _compute_filename_and_url(filename=None, prefix='altair',
                         base_url='/', nbserver_cwd='~',
                         ext='.csv'):

    if filename is None:
        filename = _compute_uuid_filename(prefix, ext)
    else:
        if not filename.endswith(ext):
            filename = filename + ext
    url = _compute_nbserver_url(filename, base_url=base_url, nbserver_cwd=nbserver_cwd)
    return filename, url


# TODO: add some tests of CodeGen
from ..codegen import CodeGen


def test_args():
    code = CodeGen('Foo', ['x', 'y', 'z'])
    assert str(code) == 'Foo(x, y, z)'


def test_kwargs():
    code = CodeGen('Bar', kwargs={'a':4, 'b':5, 'c':6})
    assert str(code) == '\n'.join(["Bar(",
                                   "    a=4,",
                                   "    b=5,",
                                   "    c=6,",
                                   ")"])


def test_code_as_arg():
    code = CodeGen('Foo', ['x', 'y', 'z'])
    code2 = CodeGen('Bar', kwargs=dict(foo=code))
    assert str(code2) == '\n'.join(['Bar(',
                                    '    foo=Foo(x, y, z),',
                                    ')'])


def test_methods():
    bar = CodeGen('bar', [1, 2, 3])
    code = CodeGen('Foo', 'x', methods=[bar])
    assert str(code) == 'Foo(x).bar(1, 2, 3)'



import json
from ..html import to_html


def test_to_html():
    json_dict = dict(A=1, B={'C':2, 'D':3})
    title = 'My Awesome Title'

    # Smoke-test for the basic template
    html = to_html(json_dict, title=title)
    assert '<title>{0}</title>'.format(title) in html

    # Test a custom template
    custom_template = "{spec}<@>{title}"
    html = to_html(json_dict, title=title, template=custom_template)
    spec, embedded_title = html.split('<@>')
    assert json.loads(spec) == json_dict
    assert embedded_title == title



"""
Histogram
-----------------
This example shows how to make a basic histogram, based on the vega-lite docs
https://vega.github.io/vega-lite/examples/histogram.html
"""
# category: basic charts

import altair as alt
from vega_datasets import data

movies = data.movies.url

chart = alt.Chart(movies).mark_bar().encode(
    alt.X("IMDB_Rating:Q", bin=True),
    y='count(*):Q',
)



"""
Choropleth Map
==============
A choropleth map of unemployment rate per county in the US
"""
# category: geographic

import altair as alt
from vega_datasets import data

counties = alt.topo_feature(data.us_10m.url,'counties')
unemp_data = alt.UrlData(data.unemployment.url)


chart = alt.Chart(counties).mark_geoshape().properties(
    projection={'type': 'albersUsa'},
    width=500,
    height=300
).encode(
    color='rate:Q'
).transform_lookup(
    lookup='id',
    from_=alt.LookupData(unemp_data, 'id', ['rate'])
)
